"""This module creates all the NamedTuple tupes of models and views
used to create figures.
"""

from typing import NamedTuple
from pathlib import Path


# class Resource(NamedTuple):
#     """The path to a file and folder."""
#
#     file: str  # The file name, with extension, with a path
#     folder: Path  # The full path that locates the file


class BaseFigure(NamedTuple):
    """The base type for all figures."""

    details: bool
    display: bool
    dpi: int
    file_stem: str
    file_type: str
    folder: Path
    latex: bool
    save: bool
    size: list[float] | None
    title: str
    xlabel: str
    xlim: list[float] | None
    ylabel: str
    ylim: list[float] | None


# class BaseModel(NamedTuple):
#     """The base type for all models."""
#
#
# class Base(NamedTuple):
#     """The base type with a base model and figure."""
#
#     figure: BaseFigure
#     model: BaseModel


# class Kwargs(NamedTuple):
#     """The keyword arguments, kwargs, forwarded to Matplotlib."""
#
#     alpha: float = 0.5
#     color: str = "blue"
#     label: str = "label"
#     linestyle: str = "-"
#     linewidth: int = 2


# class Model(NamedTuple):
#     """The basic ingredients to create a single (x, y) trace."""
#
#     folder: Path
#     file: Path
#     plot_kwargs: Kwargs
#     skip_rows: int = 1
#     ycolumn: int = 1


# class RhsAxis(NamedTuple):
#     """The constituents of a right hand side vertical axis."""
#
#     scale: float
#     label: str
#     yticks: list[float] | None


# class Figure(NamedTuple):
#    """The plot on which the models are placed."""
#
#    folder: Path
#    file: Path
#    size: list[float]
#    title: str
#    ylabel: str
#    yaxis_rhs: RhsAxis | None
#    details: bool = False
#    display: bool = True
#    dpi: int = 100
#    latex: bool = False
#    serialize: bool = False



"""This module creates a dictionary from a yml file."""

from pathlib import Path

import yaml


def yml_to_dict(fin: Path) -> dict:
    """Gien a path to a yml file, create a dict."""

    # Compared to the lower() method, the casefold() method is stronger.
    # It will convert more characters into lower case, and will find more
    # matches on comparison of two strings that are both are converted
    # using the casefold() method.
    file_type = fin.suffix.casefold()

    supported_types = (".yaml", ".yml")

    if file_type not in supported_types:
        raise TypeError("Only file types .yaml, and .yml are supported.")

    try:
        with open(file=fin, mode="r", encoding="utf-8") as stream:
            # See deprecation warning for plain yaml.load(input) at
            # https://github.com/yaml/pyyaml/wiki/PyYAML-yaml.load(input)-Deprecation
            db = yaml.load(stream, Loader=yaml.SafeLoader)
    except yaml.YAMLError as error:
        print(f"Error with YAML file: {error}")
        # print(f"Could not open: {self.self.path_file_in}")
        print(f"Could not open or decode: {fin}")
        # raise yaml.YAMLError
        raise OSError from error

    return db



"""This module is the factory, creator of models and views."""

from figio.xymodel import XYModel, XYModelAbaqus
from figio.xyview import XYView, XYViewAbaqus

# Figure Factory
FACTORY_ITEMS = {
    "model": XYModel,
    "view": XYView,
    "model_abaqus": XYModelAbaqus,
    "view_abaqus": XYViewAbaqus,
}


class XYFactory:
    """The one and only (singleton) factory for XY items."""

    @staticmethod
    def create(item, **kwargs):
        "Main factory method, returns XY objects."
        if item.startswith("xymodel"):
            instance = XYModel
        elif item.startswith("xyfigure"):
            instance = XYView
        else:
            instance = FACTORY_ITEMS.get(kwargs["class"], None)
        if instance:
            return instance(item, **kwargs)

        # If we get here, we did not return an instance, so warn.
        print(f"Warning: key 'class' specified uknown 'value' of '{item}'.")
        print("This key will be skipped.")
        return None



import os
from pathlib import Path
import sys
from abc import ABC


def absolute_path(folder):
    """
    Makes certain the path to the folder for pending serialization exists.
    If it doesn't exist, ask the user if the folder should be created or not.
    Print the full path to the command line.
    Returns the absolute path, possibly for pending serialization.
    """
    run_path = Path.cwd()
    abs_path = run_path.joinpath(folder)
    # abs_path = os.path.join(os.getcwd(), folder)
    if not os.path.isdir(abs_path):
        print(f'Folder needed but not found: "{abs_path}"')
        val = input("Create folder? [y]es or [n]o : ")
        if val == "y":
            os.mkdir(folder)
            print(f'Created folder: "{folder}"')
        else:
            print("Check accuracy of folders in database.")
            print("Abnormal script termination.")
            sys.exit("Folder misspecified.")
    return abs_path


class XYBase(ABC):
    """
    Base class to collect all data and methods common to XYBase descendants.
    """

    def __init__(self, guid, **kwargs):

        self._guid = guid

        self._verbose = kwargs.get("verbose", True)

        # moved up from XYView
        self._serialize = kwargs.get("serialize", False)

        default_folder = "."
        self._folder = kwargs.get("folder", default_folder)
        self._folder_pathlib = Path(self._folder).expanduser()
        if not self._folder_pathlib.is_dir():
            print('Error: keyword "folder" has a value (e.g., a folder path)')
            print("that cannot be found as specified:")
            print(self._folder_pathlib)
            raise KeyError("folder not found")

        self._file = kwargs.get("file", None)

        if self._file is None:
            print('Error: keyword "file" not found.')
            sys.exit("Abnormal termination.")

        # abs_path = absolute_path(self._folder)

        # self._path_file_input = os.path.join(abs_path, self._file)
        self._file_pathlib = self._folder_pathlib.joinpath(self._file)
        self._path_file_input = str(self._file_pathlib)
        self._path_file_output = None

    @property
    def guid(self):
        """Returns a globally unique identifier."""
        return self._guid

    def serialize(self, folder, filename):
        """
        Writes our data, to be extended by descendants.
        """
        abs_path = absolute_path(folder)
        self._path_file_output = os.path.join(abs_path, filename)



"""A composite is a single figure with one or more models."""

from typing import NamedTuple

import matplotlib.pyplot as plt

from figio import figure, histogram, utility


class Composite(NamedTuple):
    """The view type."""

    figure: figure.Figure
    models: list[histogram.Histogram]


def plot_composite(cc: Composite) -> None:
    """Plot the figure and histogram."""

    ff = cc.figure
    hists = cc.models

    # breakpoint()
    # Create the histogram
    fig, ax = plt.subplots(dpi=ff.dpi)
    # plt.hist(data, bins=20, color="blue", alpha=0.7, log=True)

    for hh in hists:
        print(f"Number of elements in histgram: {len(hh.data)}")
        # n_neg = [x <= 0.0 for x in bb]
        ax.hist(
            hh.data,
            **hh.plot_kwargs,
        )

    # Add a legend
    plt.legend()

    # Add title and labels
    # ax.set_title(ff.title)
    fig.suptitle(ff.title)
    ax.set_xlabel(ff.xlabel)
    ax.set_ylabel(ff.ylabel)

    if ff.details:
        ss = ff.file.name
        details = utility.timestamp(figure_name=ss)
        ax.set_title(details, fontsize=8, ha="center", color="dimgray")

    # plt.xticks(np.arange(-0.1, 1.0, 0.1))
    # xt = [-0.25, 0.0, 0.25, 0.5, 0.75, 1.00]
    # plt.xticks(xt)
    # plt.xlim([xt[0], xt[-1]])
    # plt.ylim([1, 2.0e6])

    # x_ticks = list(range(nxp))
    # y_ticks = list(range(nyp))
    # z_ticks = list(range(nzp))

    # ax.set_xlim(float(x_ticks[0]), float(x_ticks[-1]))

    # Save the plot
    if ff.save:
        # fn = Path(ff.file).stem + "_msj" + ".png"
        fn = ff.file
        plt.savefig(fn)
        print(f"Saved file: {fn}")

    # Show the plot
    if ff.display:
        plt.show()

    # Clear the current figure
    # plt.clf()



import sys

import numpy as np
from pathlib import Path
from scipy import signal

# from scipy.integrate import cumtrapz  # version 0.14.0, now deprecated
from scipy.integrate import cumulative_trapezoid  # version 1.14.1


# local application/library specific imports
# from xyfigure.xybase import XYBase
# from xyfigure.code.xybase import XYBase
# from xyfigure.xybase import XYBase, absolute_path
from figio.xybase import XYBase


def cross_correlation(reference, subject, verbose=False):
    """Creates a cross correlation assessment of how close the subject signal
    is to the reference signal.
    """

    if verbose:
        print("\nThis is xymodel.cross_correlation...")
        print(f"reference: {reference}")
        print(f"subject: {subject}")

    ref_delta_t = reference[1, 0] - reference[0, 0]
    ref_t_min = reference[0, 0]
    ref_t_max = reference[-1, 0]

    dt = subject[1, 0] - subject[0, 0]  # sample delta t
    t_min = subject[0, 0]
    t_max = subject[-1, 0]

    # globalized interval and frequency
    DT = np.minimum(ref_delta_t, dt)
    T_MIN = np.minimum(ref_t_min, t_min)
    T_MAX = np.maximum(ref_t_max, t_max)

    n_samples = int((T_MAX - T_MIN) / DT) + 1
    t_span = np.linspace(T_MIN, T_MAX, n_samples, endpoint=True)

    if verbose:
        print("\nSynchronization...")
        print(
            f"  Reference [t_min, t_max] by dt (s): [{ref_t_min}, {ref_t_max}] by {ref_delta_t}"
        )
        print(
            f"  Subject [t_min, t_max] by dt (s): [{t_min}, {t_max}] by {dt}"
        )
        print(
            f"  Globalized [t_min, t_max] by dt (s): [{T_MIN}, {T_MAX}] by {DT}"
        )
        print(f"  Globalized times: {t_span}")
        print(f"  Length of globalized times: {len(t_span)}")

    ref_y_span = np.interp(
        t_span, reference[:, 0], reference[:, 1], left=0.0, right=0.0
    )

    y_span = np.interp(
        t_span, subject[:, 0], subject[:, 1], left=0.0, right=0.0
    )

    cross_corr = np.correlate(ref_y_span, y_span, mode="full")
    cross_corr_max = np.max(cross_corr)

    cross_corr_unit = np.correlate(
        ref_y_span / np.linalg.norm(ref_y_span),
        y_span / np.linalg.norm(y_span),
        mode="full",
    )

    ref_self_corr = np.correlate(ref_y_span, ref_y_span)[
        0
    ]  # self correlated reference
    rel_corr_error = 0.0

    if ref_self_corr > 0:
        rel_corr_error = abs(cross_corr_max - ref_self_corr) / ref_self_corr

    offset_index = np.argmax(cross_corr)

    # shift time full-left, then incrementally to the right
    # t_shift = t_span - t_span[-1] + t_span[offset_index]  # nope!
    # t_shift = t_span - t_span[-1] + offset_index * DT  # bug! should shift to t0 referance signal
    t_shift = t_span - t_span[-1] + t_span[0] + offset_index * DT

    T_MIN_CORR = np.minimum(ref_t_min, t_shift[0])
    T_MAX_CORR = np.maximum(ref_t_max, t_shift[-1])

    n_samples_corr = (
        int((T_MAX_CORR - T_MIN_CORR) / DT) + 1
    )  # DT unchanged pre-shift
    t_span_corr = np.linspace(
        T_MIN_CORR, T_MAX_CORR, n_samples_corr, endpoint=True
    )

    ref_y_span_corr = np.interp(
        t_span_corr, reference[:, 0], reference[:, 1], left=0.0, right=0.0
    )

    y_span_corr = np.interp(t_span_corr, t_shift, y_span, left=0.0, right=0.0)

    error = y_span_corr - ref_y_span_corr

    L2_norm_error_rate = np.linalg.norm(error) / n_samples_corr

    if verbose:
        print("\nCorrelation...")
        print(f"  Sliding dot product (cross-correlation): {cross_corr}")
        print(f"  Length of the sliding dot product: {len(cross_corr)}")
        print(
            f"  Max sliding dot product (cross-correlation): {cross_corr_max}"
        )
        print(
            f"  Sliding dot product of normalized signals (cross-correlation): {cross_corr_unit}"
        )

        print(f"  Correlated time_shift (from full left)={offset_index * DT}")
        print(f"  Correlated index_shift (from full left)={offset_index}")

        print(f"  Correlated time step (s): {DT}")
        print(f"  Correlated t_min (s): {T_MIN_CORR}")
        print(f"  Correlated t_max (s): {T_MAX_CORR}")
        print(f"  Correlated times: {t_span_corr}")
        print(f"  Correlated reference f(t): {ref_y_span_corr}")
        print(f"  Correlated subject f(t): {y_span_corr}")
        print(f"  Correlated error f(t): {error}")

        print(f"  reference_self_correlation: {ref_self_corr}")
        print(f"  cross_correlation: {cross_corr_max}")
        print(f"    >> cross_correlation_relative_error={rel_corr_error}")
        print(f"    >> L2-norm error rate: {L2_norm_error_rate}")

    return t_span_corr, y_span_corr, rel_corr_error, L2_norm_error_rate


class XYModel(XYBase):
    """The data to be plotted in XY format."""

    def __init__(self, guid, **kwargs):
        super().__init__(guid, **kwargs)

        # TODO: rearchitect into single parent for both XYModel and
        # XYModelAbaqus
        # make sure models have an input file that exists
        if not self._file_pathlib.is_file():
            print('Error: keyword "file" has a value (e.g., a file name):')
            print(self._file)
            print("with full path specification:")
            print(self._file_pathlib)
            raise KeyError("file not found")

        self._skip_rows = kwargs.get("skip_rows", 0)
        self._skip_rows_footer = kwargs.get("skip_rows_footer", 0)
        self._xcolumn = kwargs.get("xcolumn", 0)  # default to the 1st column
        self._ycolumn = kwargs.get("ycolumn", 1)  # default to the 2nd column

        # relative to current run location
        # rel_path_and_file = os.path.join(self._folder, self._file)
        # self._data = np.genfromtxt(
        #     rel_path_and_file,
        #     dtype="float",
        #     delimiter=",",
        #     skip_header=self._skip_rows,
        #     skip_footer=self._skip_rows_footer,
        #     usecols=(self._xcolumn, self._ycolumn),
        # )
        self._data = np.genfromtxt(
            self._path_file_input,
            dtype="float",
            delimiter=",",
            skip_header=self._skip_rows,
            skip_footer=self._skip_rows_footer,
            usecols=(self._xcolumn, self._ycolumn),
        )

        # default value if plot_kwargs not client-supplied
        default = {"linewidth": 2.0, "linestyle": "-"}
        self._plot_kwargs = kwargs.get("plot_kwargs", default)

        self._xscale = kwargs.get("xscale", 1.0)
        self._yscale = kwargs.get("yscale", 1.0)
        self._xoffset = kwargs.get("xoffset", 0.0)
        self._yoffset = kwargs.get("yoffset", 0.0)

        self._signal_process = kwargs.get("signal_process", None)

        if self._signal_process:
            for process_id in self._signal_process:
                process_dict = self._signal_process.get(process_id)
                key = next(iter(process_dict))
                value = process_dict[key]

                try:
                    method = getattr(self, key)
                    method(value)
                except AttributeError as error:
                    print(f"Error: invalid signal process key: {key}")
                    print(error.__class__.__name__)

                if self._verbose:
                    print('  Signal process "' + key + '" completed.')

    @property
    def x(self):
        """Returns the model's x data."""
        return self._data[:, 0] * self._xscale + self._xoffset

    @property
    def y(self):
        """Returns the model's y data."""
        return self._data[:, 1] * self._yscale + self._yoffset

    @property
    def plot_kwargs(self):
        """Returns kwargs passed to matplotlib.pyplot.plot()."""
        return self._plot_kwargs

    def serialize(self, folder, filename, header=""):  # extend base class
        super().serialize(folder, filename)
        np.savetxt(
            self._path_file_output,
            np.transpose([self._data[:, 0], self._data[:, 1]]),
            delimiter=",",
            header=header,
        )
        if self._verbose:
            print(f"  Serialized model to: {self._path_file_output}")

    def butterworth(self, value):

        fc = value.get("cutoff", None)  # Hz, cutoff frequency
        if fc is None:
            print('Error: keyword "cutoff" not found.')
            sys.exit("Abnormal termination.")

        filter_order = value.get("order", None)
        if filter_order is None:
            print('Error: keyword "order" not found.')
            sys.exit("Abnormal termination.")

        filter_type = value.get("type", None)
        if filter_type is None:
            print('Error: keyword "type" not found.')
            sys.exit("Abnormal termination.")

        if self._verbose:
            print('Signal process "butterworth" with:')
            print(f"  cutoff frequency = {fc} Hz")
            print(f"  filter order = {filter_order}")
            print("  filter type = " + filter_type)

        dt = self._data[1, 0] - self._data[0, 0]  # sample delta t

        if self._verbose:
            print(f"  sample delta t = {dt} seconds")

        fs = 1.0 / dt  # Hz
        if self._verbose:
            print(f"  sample frequency = {fs} Hz")

        Wn = fc / (fs / 2)  # normalized critical frequency
        if self._verbose:
            print(f"  normalized critical frequency = {Wn} Hz/Hz")

        b, a = signal.butter(filter_order, Wn, filter_type)
        yfiltered = signal.filtfilt(b, a, self._data[:, 1])
        self._data[:, 1] = yfiltered  # overwrite

        serialize = value.get("serialize", 0)  # default is not to serialize
        if serialize:
            folder = value.get("folder", ".")  # default to current folder
            file_output = value.get("file", None)

            if file_output is None:
                print('Error: keyword "file" not found.')
                sys.exit("Abnormal termination.")
            else:
                self.serialize(folder, file_output)

    def correlation(self, value):

        reference = value.get("reference", None)
        if reference is None:
            print('Error: keyword "reference" not found.')
            sys.exit("Abnormal termination.")

        # ref is the reference signal to compare with
        ref_default_folder = "."
        ref_folder = reference.get("folder", ref_default_folder)

        ref_file = reference.get("file", None)

        verbosity = value.get("verbose", False)

        if ref_file is None:
            print('Error: keyword "file" not found.')
            sys.exit("Abnormal termination.")

        # abs_path = absolute_path(ref_folder)
        abs_path = Path(ref_folder).expanduser()
        # ref_path_file_input = os.path.join(abs_path, ref_file)
        ref_path_file_input = abs_path.joinpath(ref_file)

        ref_skip_rows = reference.get("skip_rows", 0)
        ref_skip_rows_footer = reference.get("skip_rows_footer", 0)
        ref_xcolumn = reference.get("xcolumn", 0)  # default to the 1st column
        ref_ycolumn = reference.get("ycolumn", 1)  # default to the 2nd column

        ref_data = np.genfromtxt(
            ref_path_file_input,
            dtype="float",
            delimiter=",",
            skip_header=ref_skip_rows,
            skip_footer=ref_skip_rows_footer,
            usecols=(ref_xcolumn, ref_ycolumn),
        )

        tcorrelated, ycorrelated, cc_relative_error, L2_error = (
            cross_correlation(ref_data, self._data, verbose=verbosity)
        )

        self._data = np.transpose([tcorrelated, ycorrelated])  # overwrite

        serialize = value.get("serialize", 0)  # default is not to serialize
        if serialize:
            folder = value.get("folder", ".")  # default to current folder
            file_output = value.get("file", None)

            if file_output is None:
                print('Error: keyword "file" not found.')
                sys.exit("Abnormal termination.")
            else:
                # self.serialize(folder, file_output, header=corr_str)
                self.serialize(folder, file_output)

    def gradient(self, value):

        gradient_order = value.get("order", None)
        if gradient_order is None:
            print('Error: keyword "order" not found.')
            sys.exit("Abnormal termination.")

        if self._verbose:
            print(
                "Signal process:"
                "  gradient applied " + str(gradient_order) + " time(s)"
            )
        # numerical gradient
        # https://docs.scipy.org/doc/numpy/reference/generated/numpy.gradient.html
        # for k in range(value):
        for k in range(gradient_order):
            ydot = np.gradient(
                self._data[:, 1], self._data[:, 0], edge_order=2
            )
            self._data[:, 1] = ydot  # overwrite
            if self._verbose:
                print(f"  Derivative {k + 1} completed.")

        serialize = value.get("serialize", 0)  # default is not to serialize
        if serialize:
            folder = value.get("folder", ".")  # default to current folder
            file_output = value.get("file", None)

            if file_output is None:
                print('Error: keyword "file" not found.')
                sys.exit("Abnormal termination.")
            else:
                self.serialize(folder, file_output)

    def integration(self, value):

        integration_order = value.get("order", None)
        if integration_order is None:
            print('Error: keyword "order" not found.')
            sys.exit("Abnormal termination.")

        print(
            'Signal process: "integration" applied '
            + str(integration_order)
            + " time(s) with"
        )

        default_ics = np.zeros(integration_order)
        ics = value.get("initial_conditions", default_ics)

        print(f"initial condition(s) as {ics}")

        if len(ics) != integration_order:
            print("Error: length of initial condition(s) array")
            print("must be equal to the order of integration.")
            print(f"Specified order = {integration_order}")
            print(f"Length of intial condition(s) array = {len(ics)}")
            sys.exit("Abnormal termination.")

        # https://docs.scipy.org/doc/numpy/reference/generated/numpy.trapz.html
        # https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.integrate.cumtrapz.html
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.cumulative_trapezoid.html#scipy.integrate.cumulative_trapezoid
        for k in range(integration_order):
            # inty = np.trapz(self._data[:, 1], self._data[:, 0]) + ics[k]
            # inty = cumtrapz(self._data[:, 1], self._data[:, 0], initial=ics[k])
            # inty = (
            #     cumtrapz(self._data[:, 1], self._data[:, 0], initial=0)
            #     + ics[k]
            # )
            inty = (
                cumulative_trapezoid(
                    self._data[:, 1], self._data[:, 0], initial=0
                )
                + ics[k]
            )

            self._data[:, 1] = inty  # overwrite
            print(f"  Integral {k + 1} completed.")

        serialize = value.get("serialize", 0)  # default is not to serialize
        if serialize:
            folder = value.get("folder", ".")  # default to current folder
            file_output = value.get("file", None)

            if file_output is None:
                print('Error: keyword "file" not found.')
                sys.exit("Abnormal termination.")
            else:
                self.serialize(folder, file_output)


class XYModelAbaqus(XYBase):
    """The ABAQUS mesh data to be plotted in XY format."""

    def __init__(self, guid, **kwargs):
        super().__init__(guid, **kwargs)

        # TODO: rearchitect into single parent for both XYModel and XYModelAbaqus
        # make sure models have an input file that exists
        if not self._file_pathlib.is_file():
            print('Error: keyword "file" has a value (e.g., a file name):')
            print(self._file)
            print("with full path specification:")
            print(self._file_pathlib)
            raise KeyError("file not found")

        with open(str(self._file_pathlib), "rt") as f:
            self._nodes = tuple()
            self._elements = tuple()
            # nice ref: https://www.pythontutorial.net/python-basics/python-read-text-file/
            try:
                # lines = f.readlines()

                # for line in f:
                line = f.readline()

                while line:
                    if "*NODE" in line or "*Node" in line:
                        # collect all nodes
                        line = f.readline()  # get the next line
                        while "*" not in line:
                            line = line.split(",")
                            # assume a 3D format even for planar problems, and catch
                            # exceptions as needed
                            try:
                                new_nodes = (
                                    tuple(
                                        [
                                            float(eval(line[1])),
                                            float(eval(line[2])),
                                            float(eval(line[3])),
                                        ]
                                    ),
                                )
                            except (
                                IndexError
                            ):  # handle 2D input files, append 0.0 as z coordinate
                                new_nodes = (
                                    tuple(
                                        [
                                            float(eval(line[1])),
                                            float(eval(line[2])),
                                            0.0,
                                        ]
                                    ),
                                )

                            self._nodes = self._nodes + new_nodes
                            # print(self._nodes)
                            line = f.readline()
                        # print(line)
                    elif "*ELEMENT" in line or "*Element" in line:
                        # collect all elements
                        line = f.readline()  # get the next line
                        while "*" not in line and len(line) > 0:
                            line = line.split(",")
                            new_element = (
                                tuple(
                                    [
                                        int(line[1]),
                                        int(line[2]),
                                        int(line[3]),
                                        int(line[4]),
                                    ]
                                ),
                            )
                            self._elements = self._elements + new_element
                            # print(self._elements)
                            line = f.readline()
                    else:
                        line = f.readline()

                print(f"Finished reading file: {self._file_pathlib}")

            except OSError:
                print(f"Cannot read file: {self._file_pathlib}")

        # default value if plot_kwargs not client-supplied
        # default = {"linewidth": 2.0, "linestyle": "solid", "color": "black", }
        _default = {
            "alpha": 0.8,
            "edgecolor": "navy",
            "facecolor": "gray",
            "linestyle": "solid",
            "linewidth": 1.0,
        }
        self._plot_kwargs = kwargs.get("plot_kwargs", _default)
        self._alpha = self._plot_kwargs.get("alpha", _default["alpha"])
        self._edgecolor = self._plot_kwargs.get(
            "edgecolor", _default["edgecolor"]
        )
        self._facecolor = self._plot_kwargs.get(
            "facecolor", _default["facecolor"]
        )
        self._linestyle = self._plot_kwargs.get(
            "linestyle", _default["linestyle"]
        )
        self._linewidth = self._plot_kwargs.get(
            "linewidth", _default["linewidth"]
        )



"""Test for the histogram type."""

import figio.histogram as histogram


def test_plot_kwargs():
    """Assures that known valid kwargs return True."""
    # Test with valid float linewidth
    valid = {"color": "blue", "linewidth": 2.0}
    assert histogram.validate_plot_kwargs(valid)

    # Test with valid integer linewidth
    valid_int = {"color": "red", "linewidth": 3}
    assert histogram.validate_plot_kwargs(valid_int)

    # Test with missing color (should be valid)
    valid_missing_color = {"linewidth": 1.5}
    assert histogram.validate_plot_kwargs(valid_missing_color)

    # Test with missing linewidth (should be valid)
    valid_missing_linewidth = {"color": "green"}
    assert histogram.validate_plot_kwargs(valid_missing_linewidth)

    # Test with negative linewidth
    invalid_negative = {"color": "blue", "linewidth": -4.0}  # Negative number
    assert not histogram.validate_plot_kwargs(invalid_negative)

    # Test with zero linewidth
    invalid_zero = {"color": "red", "linewidth": 0}  # Zero value
    assert not histogram.validate_plot_kwargs(invalid_zero)

    # Test with non-numeric linewidth
    invalid_non_numeric = {
        "color": "green",
        "linewidth": "thick",
    }  # Not a number
    assert not histogram.validate_plot_kwargs(invalid_non_numeric)

    # Test with extra keys (should still be valid)
    valid_extra_keys = {
        "color": "blue",
        "linewidth": 2.0,
        "extra_key": "value",
    }
    assert histogram.validate_plot_kwargs(valid_extra_keys)



"""This module parses yml files into types."""

from pathlib import Path


# from schema import Schema, And, Or, Use, Optional, SchemaError
from schema import Schema, And, Or, SchemaError
import yaml


# from figio.xytypes import Resource
#
## Define the schema for the 'resource' key
# resource_schema = Schema(
#    Or(
#        {
#            "folder": And(Path, len),  # "folder" must be a non-empty Path
#            "file": And(str, len),  # "file" must be a non-empty string
#        },
#        None,
#    )
# )
#
## Define the main schema that includes the "base" key
# main_schema = Schema(
#    {
#        "resource": resource_schema,
#    },
#    ignore_extra_keys=True,  # Allow any additional keys
# )


def parse_model(input: dict) -> Model:
    """Parse the input dictionary into a Resource object."""
    # Validate the input data
    validated_data = resource_schema.validate(input)

    # Extract the "resource" data
    base_data = validated_data["resource"]

    # Create a Resource object
    if base_data is not None:
        return Resource(
            folder=Path(base_data["folder"]).expanduser(),
            file=Path(base_data["folder"])
            .expanduser()
            .joinpath(base_data["file"]),
        )
    else:
        return None


# Example data to validate
vd0 = {
    "base": {
        "folder": "~/autotwin/figio/tests/input",
        "file": "base.yml",
    }
}

vd1 = {"base": None}

fa = Path(__file__).parent.joinpath("displacement_sr2_new.yml")
assert fa.is_file()
with open(str(fa), "r") as file:
    da = yaml.safe_load(file)

# Validate the example data
for item in (vd0, vd1, da):
    try:
        validated_data = main_schema.validate(item)
        print(f"Valid: Validated data for {item}:\n\n", validated_data)
    except SchemaError as e:
        print(f"Error: Validation error for {item}:\n\n", e)

# for item in (vd0, vd1, da):
for item in (da,):
    if item.get("base") is not None:
        print("base is not None")
        base = Resource(
            folder=Path(item["base"]["folder"]).expanduser(),
            file=Path(item["base"]["folder"])
            .expanduser()
            .joinpath(item["base"]["file"]),
        )
        print(f"base.folder = {base.folder}")
        print(f"base.file = {base.file}")
        # breakpoint()
        assert base.folder.is_dir(), f"Error: folder not found {base.folder}"
        assert base.file.is_file(), f"Error: file not found {base.file}"
    else:
        print("base is None")

# breakpoint()


# fb = Path(__file__).parent.joinpath("displacement_sr2_new.yml")
# assert fb.is_file()
# with open(str(fb), "r") as file:
#    db = yaml.safe_load(file)

aa = 4



import os
from datetime import datetime
from pathlib import Path
import socket

# related third-party imports
import matplotlib.pyplot as plt
from PIL import Image
from tzlocal import get_localzone

# local application/library specific imports
# from xyfigure.xybase import XYBase
# from xyfigure.code.xybase import XYBase
from figio.xybase import XYBase


class XYViewBase(XYBase):
    """The base class used by all XYViews."""

    def __init__(self, guid, **kwargs):
        super().__init__(guid, **kwargs)
        self._models = []
        self._model_keys = kwargs.get("model_keys", None)
        self._figure = None
        # self._folder = kwargs.get('folder', None)
        self._file_base = self._file.split(".")[0]

        self._title = kwargs.get("title", "default title")
        self._xlabel = kwargs.get("xlabel", "default x axis label")
        self._ylabel = kwargs.get("ylabel", "default y axis label")

        self._xticks = kwargs.get("xticks", None)
        self._yticks = kwargs.get("yticks", None)

        self._size = kwargs.get("size", [11.0, 8.5])
        self._dpi = kwargs.get("dpi", 100)
        self._xlim = kwargs.get("xlim", None)
        self._ylim = kwargs.get("ylim", None)

        self._display = kwargs.get("display", True)
        self._details = kwargs.get("details", True)
        # self._serialize = kwargs.get('serialize', False) # moved up to XYBase
        self._latex = kwargs.get("latex", False)
        if self._latex:
            from matplotlib import rc

            rc("text", usetex=True)
            rc("font", family="serif")

        # default value if axes_kwargs not client-supplied
        # https://matplotlib.org/stable/api/axes_api.html
        # https://matplotlib.org/stable/api/axes_api.html#appearance
        # whether frame (rectangle outline) appears on all axes
        self._frame = kwargs.get("frame", True)

        # https://matplotlib.org/stable/gallery/subplots_axes_and_figures/axes_props.html#sphx-glr-gallery-subplots-axes-and-figures-axes-props-py
        #
        # https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.tick_params.html#matplotlib.axes.Axes.tick_params
        # default to empty dictionary
        self._tick_params = kwargs.get("tick_params", {})

        print("Finished XYViewBase constructor.")

    @property
    def models(self):
        return self._models

    @models.setter
    def models(self, value):
        self._models = value

    @property
    def model_keys(self):
        return self._model_keys

    def figure(self):
        pass

    def serialize(self, folder, filename):  # extend base class
        # deprecated
        # super().serialize(folder, filename)
        # self._figure.savefig(
        #     self._path_file_output, dpi=self._dpi, bbox_inches="tight"
        # )  # avoid cutoff of labels
        # print(f"  Serialized view to: {self._path_file_output}")

        # New for XYView class, the so-called input file (input by the user) is the
        # output file to which the figure should be serialized.
        self._figure.savefig(
            self._path_file_input, dpi=self._dpi, bbox_inches="tight"
        )  # avoid cutoff of labels
        print(f"  Serialized view to: {self._path_file_input}")


class XYView(XYViewBase):
    """Creates a view that sees models."""

    def __init__(self, guid, **kwargs):
        super().__init__(guid, **kwargs)
        # self._models = []
        # self._model_keys = kwargs.get("model_keys", None)
        # self._figure = None
        # # self._folder = kwargs.get('folder', None)
        # self._file_base = self._file.split(".")[0]

        # default value if figure_kwargs not client-supplied
        # self._title = kwargs.get("title", "default title")
        # self._xlabel = kwargs.get("xlabel", "default x axis label")
        # self._ylabel = kwargs.get("ylabel", "default y axis label")

        # self._xticks = kwargs.get("xticks", None)
        # self._yticks = kwargs.get("yticks", None)

        self._xaxislog = kwargs.get("xaxis_log", False)
        self._yaxislog = kwargs.get("yaxis_log", False)

        # default = {'scale': 1, 'label': 'ylabel_rhs', 'verification': 0}
        self._yaxis_rhs = kwargs.get("yaxis_rhs", None)

        # inches, U.S. paper, landscape
        # self._size = kwargs.get("size", [11.0, 8.5])
        # self._dpi = kwargs.get("dpi", 100)
        # self._xlim = kwargs.get("xlim", None)
        # self._ylim = kwargs.get("ylim", None)

        self._background_image = kwargs.get("background_image", None)

        # self._display = kwargs.get("display", True)
        # self._details = kwargs.get("details", True)
        # # self._serialize = kwargs.get('serialize', False) # moved up to XYBase
        # self._latex = kwargs.get("latex", False)
        # if self._latex:
        #     from matplotlib import rc

        #     rc("text", usetex=True)
        #     rc("font", family="serif")

    # @property
    # def models(self):
    #     return self._models

    # @models.setter
    # def models(self, value):
    #     self._models = value

    # @property
    # def model_keys(self):
    #     return self._model_keys

    def figure(self):
        """Create a figure (view) of the registered models to the screen."""
        if self._figure is None:

            # dpi versus fig size
            # https://stackoverflow.com/questions/47633546/relationship-between-dpi-and-figure-size
            # fig, ax = plt.subplots(nrows=1, dpi=self._dpi)
            self._figure, ax = plt.subplots(nrows=1, dpi=self._dpi)
            if self._verbose:
                print(f"  Figure dpi set to {self._dpi}")

            # ax.ticklabel_format(axis='y', style='scientific')
            # ax.ticklabel_format(axis='both', style='scientific', scilimits=(0,0))

            # https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.figure.Figure.html#matplotlib.figure.Figure.set_size_inches
            # fig.set_size_inches(self._size)
            self._figure.set_size_inches(self._size)
            if self._verbose:
                print("  Figure size set to " + str(self._size) + " inches.")

            if self._background_image:
                folder = self._background_image.get("folder", ".")
                file = self._background_image.get("file", None)
                # rel_path_and_file = os.path.join(
                #     folder, file
                # )  # relative to current run location
                path_file = Path(folder).expanduser().joinpath(file)
                # im = Image.open(rel_path_and_file)
                im = Image.open(path_file)

                left = self._background_image.get("left", 0.0)
                right = self._background_image.get("right", 1.0)
                bottom = self._background_image.get("bottom", 0.0)
                top = self._background_image.get("top", 1.0)
                al = self._background_image.get("alpha", 1.0)

                # https://github.com/matplotlib/matplotlib/issues/3173
                # https://matplotlib.org/3.1.1/tutorials/intermediate/imshow_extent.html
                bounds = [left, right, bottom, top]
                im = ax.imshow(
                    im, zorder=0, extent=bounds, alpha=al, aspect="auto"
                )

            for model in self._models:
                # needs rearchitecting, a logview descends from a view
                if (
                    self._xaxislog and not self._yaxislog
                ):  # needs rearchitecting
                    ax.semilogx(model.x, model.y, **model.plot_kwargs)
                elif not self._xaxislog and self._yaxislog:
                    ax.semilogy(model.x, model.y, **model.plot_kwargs)
                elif self._xaxislog and self._yaxislog:
                    ax.loglog(model.x, model.y, **model.plot_kwargs)
                else:
                    ax.plot(model.x, model.y, **model.plot_kwargs)

            if self._xticks:
                ax.set_xticks(self._xticks)

            if self._yticks:
                ax.set_yticks(self._yticks)

            if self._xlim:
                ax.set_xlim(self._xlim)

            if self._ylim:
                ax.set_ylim(self._ylim)

            if self._yaxis_rhs:
                rhs_axis_scale = self._yaxis_rhs.get("scale", 1)
                rhs_axis_label = self._yaxis_rhs.get("label", None)
                # rhs_yticks_str = self._yaxis_rhs.get('yticks', None)
                rhs_yticks = self._yaxis_rhs.get("yticks", None)

                # ax2 = fig.add_subplot(111, sharex=ax, frameon=False)
                ax2 = self._figure.add_subplot(111, sharex=ax, frameon=False)
                bottom, top = ax.get_ylim()  # get from left-hand-side y-axis
                ax2.set_ylim(rhs_axis_scale * bottom, rhs_axis_scale * top)
                ax2.yaxis.tick_right()
                # ax2.ticklabel_format(axis='both', style='scientific', scilimits=(0,0))
                # ax.ticklabel_format(axis='both', style='scientific', scilimits=(0,0))
                # _ticklabel_format = self._yaxis_rhs.get('ticklabel_format', None)
                # _ticklabel_format = self._yaxis_rhs.get('ticklabel_format', None)
                # if _ticklabel_format:
                #     scilimits_str = _ticklabel_format.get('scilimitsl', "(0, 0)")
                #     ax2.ticklabel_format(**_ticklabel_format)
                # ax2.ticklabel_format(axis='both', style='scientific', scilimits=(0,0))
                # plt.ticklabel_format(axis='y', style='scientific', useOffset=False)
                # ax2.ticklabel_format(axis='y', style='scientific', useOffset=False)
                if rhs_yticks:
                    ax2.set_yticks(rhs_yticks)
                ax2.yaxis.set_label_position("right")
                ax2.set_ylabel(rhs_axis_label)

            # fig.suptitle(self._title)
            self._figure.suptitle(self._title)
            ax.set_xlabel(self._xlabel)
            ax.set_ylabel(self._ylabel)
            # set frame on or off based on the Bool "frame" in .json input
            ax.set_frame_on(b=self._frame)
            if len(self._tick_params) > 0:
                ax.tick_params(**self._tick_params)
            ax.grid()
            ax.legend()

            if self._details:
                # Get the local timezone
                local_timezone = get_localzone()

                # Get the current time in the local timezone
                current_time = datetime.now(local_timezone)

                # Format the date and time stamp
                timestamp = current_time.strftime("%Y-%m-%d %H:%M:%S %Z%z")

                user = str(os.getlogin())
                # host = str(os.getenv("HOSTNAME"))
                host = socket.gethostname()
                details_str = (
                    self._file
                    + " created "
                    + timestamp
                    + " by "
                    + user
                    + " on "
                    + host
                )

                ax.set_title(
                    details_str, fontsize=8, ha="center", color="dimgray"
                )

            if self._display:
                plt.show()

            if self._serialize:
                self.serialize(self._folder, self._file)

            plt.close("all")
            self._figure = None

    # def serialize(self, folder, filename):  # extend base class
    #     # deprecated
    #     # super().serialize(folder, filename)
    #     # self._figure.savefig(
    #     #     self._path_file_output, dpi=self._dpi, bbox_inches="tight"
    #     # )  # avoid cutoff of labels
    #     # print(f"  Serialized view to: {self._path_file_output}")

    #     # New for XYView class, the so-called input file (input by the user) is the
    #     # output file to which the figure should be serialized.
    #     self._figure.savefig(
    #         self._path_file_input, dpi=self._dpi, bbox_inches="tight"
    #     )  # avoid cutoff of labels
    #     print(f"  Serialized view to: {self._path_file_input}")


class XYViewAbaqus(XYViewBase):
    """Creates an ABAQUS view that sees ABAQUS models."""

    def __init__(self, guid, **kwargs):
        super().__init__(guid, **kwargs)
        # self._models = []
        # self._model_keys = kwargs.get("model_keys", None)
        # self._figure = None
        # # self._folder = kwargs.get('folder', None)
        # self._file_base = self._file.split(".")[0]

        # default value if figure_kwargs not client-supplied
        # self._title = kwargs.get("title", "default title")
        # self._xlabel = kwargs.get("xlabel", "default x axis label")
        # self._ylabel = kwargs.get("ylabel", "default y axis label")

        # self._xticks = kwargs.get("xticks", None)
        # self._yticks = kwargs.get("yticks", None)

        # inches, U.S. paper, landscape
        # self._size = kwargs.get("size", [11.0, 8.5])
        # self._dpi = kwargs.get("dpi", 100)
        # self._xlim = kwargs.get("xlim", None)
        # self._ylim = kwargs.get("ylim", None)

        # self._display = kwargs.get("display", True)
        # self._details = kwargs.get("details", True)
        # # self._serialize = kwargs.get('serialize', False) # moved up to XYBase
        # self._latex = kwargs.get("latex", False)
        # if self._latex:
        #     from matplotlib import rc

        #     rc("text", usetex=True)
        #     rc("font", family="serif")

    @property
    def models(self):
        return self._models

    @models.setter
    def models(self, value):
        self._models = value

    @property
    def model_keys(self):
        return self._model_keys

    def figure(self):
        """Create a figure (view) of the registered models to the screen."""
        if self._figure is None:

            self._figure, ax = plt.subplots(nrows=1, dpi=self._dpi)
            if self._verbose:
                print(f"  Figure dpi set to {self._dpi}")

            self._figure.set_size_inches(self._size)
            if self._verbose:
                print("  Figure size set to " + str(self._size) + " inches.")

            for model in self._models:
                xs, ys, _ = zip(*model._nodes)

                for face in model._elements:
                    xf = tuple(
                        xs[k - 1] for k in face
                    )  # 1-base index to 0-base index
                    yf = tuple(ys[k - 1] for k in face)
                    # plt.fill(
                    #     xf,
                    #     yf,
                    #     linestyle="dotted",
                    #     edgecolor="magenta",
                    #     alpha=0.5,
                    #     facecolor="gray",
                    # )
                    plt.fill(
                        xf,
                        yf,
                        alpha=model._alpha,
                        edgecolor=model._edgecolor,
                        facecolor=model._facecolor,
                        linestyle=model._linestyle,
                        linewidth=model._linewidth,
                    )

            if self._xticks:
                ax.set_xticks(self._xticks)

            if self._yticks:
                ax.set_yticks(self._yticks)

            if self._xlim:
                ax.set_xlim(self._xlim)

            if self._ylim:
                ax.set_ylim(self._ylim)

            if self._xlabel:
                ax.set_xlabel(self._xlabel)

            if self._ylabel:
                ax.set_ylabel(self._ylabel)

            # set frame on or off based on the Bool "frame" in .json input
            ax.set_frame_on(b=self._frame)
            if len(self._tick_params) > 0:
                ax.tick_params(**self._tick_params)

            if self._display:
                plt.show()

            if self._serialize:
                self.serialize(self._folder, self._file)

            plt.close("all")
            self._figure = None



"""The figure type, constructor, and methods."""

from typing import NamedTuple
from pathlib import Path

import matplotlib.pyplot as plt
from schema import Schema, And, Optional, SchemaError, Use

from figio import histogram
from figio import utility

# Define the schema for the figure
figure_schema = Schema(
    {
        Optional("details", default=True): bool,
        Optional("display", default=True): bool,
        Optional("dpi", default=100): int,
        "folder": And(str, len),  # non-empty string
        "file": And(str, len),  # non-empty string
        Optional("latex", default=False): bool,
        "models": And(
            [Use(str)],  # each item in the list must be a string
            lambda lst: len(lst) >= 1,  # list must have at least 1 item
        ),
        Optional("save", default=False): bool,
        Optional("size", default=[8.0, 6.0]): list[float | int],
        Optional("title", default=""): str,
        Optional("xlabel", default=""): str,
        Optional("ylabel", default=""): str,
    },
    ignore_extra_keys=True,
)


class Figure(NamedTuple):
    """The figure type."""

    details: str
    display: bool
    dpi: int
    folder: Path
    file: Path
    latex: bool
    models: list[str]
    save: bool
    size: list[float | int]
    title: str
    xlabel: str
    ylabel: str
    # yaxis_rhs: RhsAxis | None


def validate_schema(din: dict) -> bool:
    """Determines if the input dictionary is a valid Figure schema."""

    # TODO: DRY out code, repeated here at in histogram.py
    try:
        validated_data = figure_schema.validate(din)
        print("Valid: Validated data:", validated_data)
    except SchemaError as e:
        print("Error: Validation error:", e)

    return True


def new(db: dict) -> Figure:
    """Given a dictionary, validates the data from the
    dictionary, and if valid, creates a Figure."""

    validate_schema(db)

    ff = Figure(
        details=db["details"],
        display=db["display"],
        dpi=db["dpi"],
        folder=Path(db["folder"]).expanduser(),
        file=Path(db["folder"]).expanduser().joinpath(db["file"]),
        latex=db["latex"],
        models=db["models"],
        save=db["save"],
        size=db["size"],
        title=db["title"],
        xlabel=db["xlabel"],
        ylabel=db["ylabel"],
    )

    assert ff.folder.is_dir(), f"Folder {ff.folder} does not exist."
    ext = ff.file.suffix
    file_types = (".pdf", ".png", ".svg")
    assert ext in file_types, f"File type {ext} not in {file_types}"

    # TODO: finish
    return ff


def plot_histogram(ff: Figure, hh: histogram.Histogram) -> None:
    """Plot the figure and histogram."""

    print(f"Number of elements: {len(hh.data)}")
    # n_neg = [x <= 0.0 for x in bb]

    # breakpoint()
    # Create the histogram
    fig, ax = plt.subplots(dpi=ff.dpi)
    # plt.hist(data, bins=20, color="blue", alpha=0.7, log=True)
    ax.hist(
        hh.data,
        **hh.plot_kwargs,
    )

    # Add title and labels
    # ax.set_title(ff.title)
    fig.suptitle(ff.title)
    ax.set_xlabel(ff.xlabel)
    ax.set_ylabel(ff.ylabel)

    if ff.details:
        ss = ff.file.name
        details = utility.timestamp(figure_name=ss)
        ax.set_title(details, fontsize=8, ha="center", color="dimgray")

    # plt.xticks(np.arange(-0.1, 1.0, 0.1))
    # xt = [-0.25, 0.0, 0.25, 0.5, 0.75, 1.00]
    # plt.xticks(xt)
    # plt.xlim([xt[0], xt[-1]])
    # plt.ylim([1, 2.0e6])

    # x_ticks = list(range(nxp))
    # y_ticks = list(range(nyp))
    # z_ticks = list(range(nzp))

    # ax.set_xlim(float(x_ticks[0]), float(x_ticks[-1]))

    # Save the plot
    if ff.save:
        # fn = Path(ff.file).stem + "_msj" + ".png"
        fn = ff.file
        plt.savefig(fn)
        print(f"Saved file: {fn}")

    # Show the plot
    if ff.display:
        plt.show()

    # Clear the current figure
    # plt.clf()



"""The histogram type, constructor, and methods."""

from typing import NamedTuple
from pathlib import Path

import numpy as np
from schema import Schema, And, Or, Use, Optional, SchemaError

from figio import utility

# Define the schema for plot_kwargs
plot_kwargs_schema = Schema(
    {
        Optional("alpha", default=1.0): And(
            Or(Use(float), Use(int)),  # must be a float or an int
            lambda n: n >= 0 and n <= 1,  # must be between 0 and 1
        ),  # Ensure alpha is a float between 0 and 1
        Optional("bins", default=20): And(
            Use(int), lambda n: n > 0
        ),  # Ensure bins is a positive integer
        Optional("color", default="black"): str,
        Optional("histtype", default="step"): str,
        Optional("label", default=""): str,
        Optional("linewidth", default=1.0): And(
            Or(Use(float), Use(int)),  # must be a float or an int
            lambda n: n > 0,  # must be positive
        ),  # Ensure linewidth is a positive float
        Optional("log", default=True): bool,
    },
    ignore_extra_keys=True,
)


class Histogram(NamedTuple):
    """The histogram type."""

    # bins: int
    # density: bool
    data: np.ndarray
    folder: Path
    file: Path
    guid: str
    # histtype: str
    # orientation: str
    plot_kwargs: dict
    # rwidth: float
    skip_rows: int
    # stacked: bool
    # x: list[float]
    # y: list[float]
    ycolumn: int
    # yerr: list[float] | None


def validate_plot_kwargs(din: dict) -> bool:
    """Determines if the input dictionary is a valid plot_kwargs schema."""

    # TODO: DRY out code, repeated here at in figure.py
    try:
        validated_data = plot_kwargs_schema.validate(din)
        print("Valid: Validated data:", validated_data)
        return True
    except SchemaError as e:
        print("Error: Validation error:", e)
        return False


def new(guid: str, db: dict) -> Histogram:
    """Given a globally unique identifier string and a dictionary,
    validates the data from the dictionary, and if valid,
    creates a Histogram."""

    plot_kwargs = utility.get_default_values(schema=plot_kwargs_schema)
    print(f"Histogram default plot_kwargs: {plot_kwargs}")

    if "plot_kwargs" in db:
        user_kwargs = db["plot_kwargs"]
        plot_kwargs.update(user_kwargs)
        print(f"Histogram updated plot_kwargs: {plot_kwargs}")

    validate_plot_kwargs(plot_kwargs)

    folder = Path(db["folder"]).expanduser()
    file = Path(db["folder"]).expanduser().joinpath(db["file"])
    assert folder.is_dir(), f"Folder does not exist: {folder}"
    assert file.is_file, f"File does not exist: {file}"

    # TODO: validate the entire schema, not just the kwargs below
    # then the next two checks are obsolete
    skip_rows = db["skip_rows"]
    ycolumn = db["ycolumn"]
    assert skip_rows >= 0, f"skip_rows must be >= 0: {skip_rows}"
    assert ycolumn >= 0, f"ycolumn must be >= 0: {ycolumn}"

    # load the data into the Histogram
    try:
        data = np.genfromtxt(
            str(file),
            dtype="float",
            delimiter=",",
            skip_header=db["skip_rows"],
            skip_footer=0,
            usecols=(db["ycolumn"]),
        )
    except FileNotFoundError as e:
        print(f"File not found: {e}")
    except ValueError as e:
        print(f"Value error: {e}")
    except IOError as e:
        print(f"I/O error: {e}")
    except TypeError as e:
        print(f"Type error: {e}")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")

    hh = Histogram(
        folder=Path(db["folder"]).expanduser(),
        file=Path(db["folder"]).expanduser().joinpath(db["file"]),
        guid=guid,
        data=data,
        plot_kwargs=plot_kwargs,
        skip_rows=skip_rows,
        ycolumn=ycolumn,
    )

    return hh



"""Module to define schema types."""

# from pathlib import Path
from typing import NamedTuple, Any
from schema import Schema, And, Use, Optional, SchemaError


def valid_model(din: dict) -> bool:
    """Determines if the input dictionary is a valid Model schema."""
    # Validate the example data

    # Define the schema for the Model class
    model_schema = Schema(
        {
            "folder": And(str, len),  # "folder" must be a non-empty string
            "file": And(str, len),  # "file" must be a non-empty string
            "plot_kwargs": dict,  # "plot_kwargs" must be a dictionary
            "skip_rows": And(
                Use(int), lambda n: n >= 0
            ),  # "skip_rows" must be a non-negative integer
            "ycolumn": And(
                Use(int), lambda n: n >= 0
            ),  # "ycolumn" must be a non-negative integer
        }
    )

    try:
        validated_data = model_schema.validate(din)
        print("Valid: Validated data:", validated_data)
    except SchemaError as e:
        print("Error: Validation error:", e)

    return True



"""This module, command_line.py, is the new command line entry point, which
accepts input files of type .yml.
"""

import argparse
from pathlib import Path

# import figio.constants as cc
from figio.factory import XYFactory
from figio import composite, figure, histogram
from figio.xymodel import XYModel, XYModelAbaqus
from figio.xyview import XYView, XYViewAbaqus
from figio.yml_to_dict import yml_to_dict


def command_line(fin: Path) -> bool:
    """Given a .yml file, processes it to create a figure.

    Args:
        fin: The fully pathed input file.

    Returns
        True if successful, False otherwise.
    """
    processed = False

    db = yml_to_dict(fin=fin)

    items = []  # cart of items is empty, fill from factory
    # factory = XYFactory()  # it's static!

    print("====================================")
    print("Information")
    print("For (x, y) data and time series data:")
    print("  xymodel_* items associate with xyfigure_* items.")
    print("For histogram data:")
    print("  hmodel_* items associate with hfigure_* items.")
    print("====================================")

    for item in db:

        if item.startswith("hmodel"):
            i = histogram.new(guid=item, db=db[item])
            items.append(i)

        elif item.startswith("hfigure"):
            i = figure.new(db[item])
            items.append(i)

        else:
            kwargs = db[item]
            i = XYFactory.create(item, **kwargs)
            if i:
                items.append(i)
            else:
                warn_msg = "Item is None from factory,"
                warn_msg += " nothing added to command_line items."
                print(warn_msg)

    hists = [i for i in items if isinstance(i, histogram.Histogram)]
    figures = [i for i in items if isinstance(i, figure.Figure)]

    # TODO: finish iteration
    for item in figures:

        print(f"Draw models: {item.models} on: {item.file.name}")
        models = [hh for hh in hists if hh.guid in item.models]

        # make a composite view
        # a_view = view.View(figure=item, models=view_models)
        a_comp = composite.Composite(figure=item, models=models)

        composite.plot_composite(a_comp)

    # TODO: refactor old models and views into above form
    models = [i for i in items if isinstance(i, (XYModel, XYModelAbaqus))]
    views = [i for i in items if isinstance(i, (XYView, XYViewAbaqus))]

    for view in views:
        print(f'Creating view with guid = "{view.guid}"')

        if view.model_keys:  # register only selected models with current view
            print(f"  Adding {view.model_keys} model(s) to current view.")
            view.models = [m for m in models if m.guid in view.model_keys]
            if len(view.models) == 0:
                warn_msg = "Warning:\n"
                warn_msg += "  Current view has no associated models.\n"
                warn_msg += "  Check the view 'model_keys' match with'\n"
                warn_msg += "  'xymodel_*' in the .yml file."
                print(warn_msg)
            view.figure()  # must be within this subset scope
        else:
            print("  Adding all models to current view.")
            view.models = models  # register all models with current view
            view.figure()  # must be within this subset scope

    print("====================================")
    print("End of figio execution.")

    processed = True  # overwrite
    return processed  # success if we reach this line


def main():
    """Runs the module from the command line."""
    # print(cl.BANNER)
    # print(cl.CLI_DOCS)
    parser = argparse.ArgumentParser(
        prog="figio",
        description="Generate a figure.",
        epilog="figio finished",
    )
    parser.add_argument(
        "input_file", help="the .yml recipe used to create the figure"
    )

    args = parser.parse_args()
    if args.input_file:
        aa = Path(args.input_file).expanduser()
        if aa.is_file():
            print(f"Processing file: {aa}")
            command_line(fin=aa)
        else:
            print(f"Error: could not find file: {aa}")


if __name__ == "__main__":
    main()



"""This module contains various utilities."""

import os
from datetime import datetime
import socket

from schema import Schema, Optional
from tzlocal import get_localzone


def timestamp(figure_name: str) -> str:
    """Returns the current timestamp."""

    # Get the local timezone
    local_timezone = get_localzone()

    # Get the current time in the local timezone
    current_time = datetime.now(local_timezone)

    # Format the date and time stamp
    dts = current_time.strftime("%Y-%m-%d %H:%M:%S %Z%z")

    user = str(os.getlogin())
    host = socket.gethostname()

    details = figure_name + " created " + dts + " by " + user + " on " + host

    return details


def get_default_values(schema: Schema):
    """Function to extract default values from the schema"""

    defaults = {}
    for key, _ in schema.schema.items():
        if isinstance(key, Optional):
            if hasattr(key, "default"):
                # defaults[key] = key.default
                defaults[key.key] = key.default
        # else:
        # If there are no defaults,
        # choose to add them as None or skip
        # defaults[key] = None

    return defaults



"""This modules test that schema."""

from pathlib import Path
import yaml

from figio.schema import valid_model


def test_model_schema():
    """WIP"""

    fa = Path(__file__).parent.joinpath("input", "rigid_body_ang_pos.yml")
    # breakpoint()
    assert fa.is_file()
    with open(str(fa), "r") as file:
        da = yaml.safe_load(file)

    assert valid_model(da)
    # assert not valid_model({"folder": "folder", "file": "file"})
